// print.inc
//
// Note: you must include the appropriate AoA header
// file for your system prior to including this
// source file (this source file is portable to all
// systems, so it cannot know which system you're
// using in order to include the header file itself). 

            .text
            .align  2            
            
// print:
//
// Prints the string following the bl instruction
// and then adjust the return address to point
// at the instruction following the string. Assumes
// the string (including terminating zero-byte)
// is a multiple of four bytes long (with padding
// at the end of the string, if necessary).
//
// Note: because this calls printf, we can actually
//       pass additional parameters to match % entries
//       in the string. Args 1-4 are passed in R0-R3
//       and args 5-8 are passed on the stack.
//       (Maximum of eight args to print.)
//
// Note: this is Thumb-1 code so it is portable to
// all CPUs that support the Thumb instruction set
// (including Cortex-A).

            proc    print, thumb
            
            .set    prt.ra,   36    // Offset to return adrs
            .set    prt.arg5, 40    // Offset to parameter 5
            .set    prt.arg6, 44    // Offset to parameter 6
            .set    prt.arg7, 48    // Offset to parameter 7
            .set    prt.arg8, 52    // Offset to parameter 8
            
            // Preserve all the volatile registers
            // so you can call print in arbitrary
            // places (good for debugging).
            
            push    {r0, r1, r2, r3, lr}
            sub     sp, #20     // Space for five parameters
            
            // Copy parameters 4 to 8 onto printf stack:
            
            str     r3, [sp]    // 4th print is 5th printf
            ldr     r3, [sp, #prt.arg5]
            str     r3, [sp, #4]
            ldr     r3, [sp, #prt.arg6]
            str     r3, [sp, #8]
            ldr     r3, [sp, #prt.arg7]
            str     r3, [sp, #12]
            ldr     r3, [sp, #prt.arg8]
            str     r3, [sp, #16]
            
            // Copy print parameters 1-3 to printf parms 2-4:
            
            mov     r3, r2
            mov     r2, r1
            mov     r1, r0
            
            // Printf format string (R0) comes from LR.
            // If this is being called from Thumb code,
            // the LO bit will be 1, so clear it:
            
            mov     r0, lr
            lsrs    r0, #1      // AND/BIC would require
            lsls    r0, #1      // using another register
            bl      printf
            
            // We need to find the end of the string
            // following the call to print. Don't forget
            // to clear bit 0 (ARM/Thumb mode bit) so
            // this code doesn't skip the first byte
            // of the string.
            
            ldr     r0, [sp, #prt.ra]   // Get LR from stack
            movs    r1, #1              // Clear Thumb/ARM
            bics    r0, r1              // state bit.
            bal     whlNot0
            .pool
            
find0lp:    adds    r0, #1
whlNot0:    ldrb    r1, [r0]
            cmp     r1, #0
            bne     find0lp
            
            // Found the zero byte. Now found R0 up
            // to the next word boundary:
            
            adds    r0, #3              // (R0 + 3) % 4
            movs    r1, #3              // Clear LO 2 bits of R0
            bics    r0, r1
            
            // Must fetch the original LR value and copy bit 0
            // (ARM/Thumb state) into bit 0 of the newly-generated
            // return address:
            
            ldr     r1, [sp, #prt.ra]
            movs    r2, #1
            ands    r1, r2              // Extract bit 0
            orrs    r0, r1              // Merge into new RA
            str     r0, [sp, #prt.ra]   // Overwrite return address
            
            add     sp, #20             // Deallocate locals
            pop     {r0, r1, r2, r3, pc}
            
            endp    print

